from collections import defaultdict
from functools import partial
from itertools import chain
from time import monotonic

from kivy.clock import Clock
from kivy.logger import Logger
from kivy.properties import (
	BooleanProperty,
	ListProperty,
	NumericProperty,
	ObjectProperty,
	ReferenceListProperty,
)
from kivy.uix.widget import Widget
from kivy.vector import Vector

from elide.boardscatter import BoardScatterPlane

from ..boardview import BoardView
from ..kivygarden.texturestack import Stack, TextureStackPlane
from ..util import store_kv


class GridPawn(Stack):
	default_image_paths = ["atlas://rltiles/base.atlas/unseen"]

	@property
	def _stack_plane(self):
		return self.board.pawn_plane


class GridSpot(Stack):
	default_image_paths = ["atlas://rltiles/floor.atlas/floor-stone"]

	@property
	def _stack_plane(self):
		return self.board.spot_plane


class GridBoard(Widget):
	selection = ObjectProperty()
	selection_candidates = ListProperty()
	character = ObjectProperty()
	tile_width = NumericProperty(32)
	tile_height = NumericProperty(32)
	tile_size = ReferenceListProperty(tile_width, tile_height)
	pawn_cls = GridPawn
	spot_cls = GridSpot

	def __init__(self, **kwargs):
		self.pawn = {}
		self.spot = {}
		self.contained = defaultdict(set)
		super().__init__(**kwargs)

	def update(self):
		make_spot = self.make_spot
		make_pawn = self.make_pawn
		self.spot_plane.data = list(
			map(make_spot, self.character.place.values())
		)
		self.pawn_plane.data = list(
			map(make_pawn, self.character.thing.values())
		)
		self.spot_plane.redraw()
		self.pawn_plane.redraw()

	def add_spot(self, placen, *args):
		if placen not in self.character.place:
			raise KeyError(f"No such place for spot: {placen}")
		if placen in self.spot:
			raise KeyError("Already have a Spot for this Place")
		spt = self.make_spot(self.character.place[placen])
		self.spot_plane.add_datum(spt)
		self.spot[placen] = self.spot_cls(board=self, proxy=spt["proxy"])

	def make_spot(self, place):
		placen = place["name"]
		if not isinstance(placen, tuple) or len(placen) != 2:
			raise TypeError(
				"Can only make spot from places with tuple names of length 2",
				placen,
			)
		if (
			not isinstance(placen, tuple)
			or len(placen) != 2
			or not isinstance(placen[0], int)
			or not isinstance(placen[1], int)
		):
			raise TypeError(
				"GridBoard can only display places named with pairs of ints"
			)
		if "_image_paths" in place:
			textures = list(place["_image_paths"])
		else:
			textures = list(self.spot_cls.default_image_paths)
		r = {
			"name": placen,
			"x": int(placen[0] * self.tile_width),
			"y": int(placen[1] * self.tile_height),
			"width": int(self.tile_width),
			"height": int(self.tile_height),
			"textures": textures,
			"proxy": place,
		}
		return r

	def make_pawn(self, thing) -> dict:
		location = self.spot[thing["location"]]
		r = {
			"name": thing["name"],
			"x": int(location.x),
			"y": int(location.y),
			"width": int(self.tile_width),
			"height": int(self.tile_height),
			"location": location,
			"textures": list(
				thing.get("_image_paths", self.pawn_cls.default_image_paths)
			),
			"proxy": thing,
		}
		return r

	def add_pawn(self, thingn, *args):
		if thingn not in self.character.thing:
			raise KeyError(f"No such thing: {thingn}")
		if thingn in self.pawn:
			raise KeyError(f"Already have a pawn for {thingn}")
		thing = self.character.thing[thingn]
		if thing["location"] not in self.spot:
			# The location is not in the grid. That's fine.
			return
		pwn = self.make_pawn(thing)
		self.pawn[thingn] = self.pawn_cls(board=self, proxy=pwn["proxy"])
		location = pwn["location"]
		self.contained[location].add(thingn)
		self.pawn_plane.add_datum(pwn)

	def _trigger_add_pawn(self, thingn):
		part = partial(self.add_pawn, thingn)
		Clock.unschedule(part)
		Clock.schedule_once(part, 0)

	def on_parent(self, *args):
		if self.character is None:
			Clock.schedule_once(self.on_parent, 0)
			return
		if self.character.engine.closed:
			return
		Logger.debug("GridBoard: on_parent start")
		start_ts = monotonic()
		if not hasattr(self, "pawn_plane"):
			self.pawn_plane = TextureStackPlane(pos=self.pos, size=self.size)
			self.spot_plane = TextureStackPlane(pos=self.pos, size=self.size)
			self.bind(
				pos=self.pawn_plane.setter("pos"),
				size=self.pawn_plane.setter("size"),
			)
			self.bind(
				pos=self.spot_plane.setter("pos"),
				size=self.spot_plane.setter("size"),
			)
			self.add_widget(self.spot_plane)
			self.add_widget(self.pawn_plane)
		spot_data = list(
			map(
				self.make_spot,
				filter(
					lambda spot: isinstance(spot["name"], tuple)
					and len(spot["name"]) == 2,
					self.character.place.values(),
				),
			)
		)
		if not spot_data:
			if hasattr(self.spot_plane, "_redraw_bind_uid"):
				self.spot_plane.unbind_uid(
					"data", self.spot_plane._redraw_bind_data_uid
				)
				del self.spot_plane._redraw_bind_data_uid
			if hasattr(self.pawn_plane, "_redraw_bind_uid"):
				self.pawn_plane.unbind_uid(
					"data", self.pawn_plane._redraw_bind_data_uid
				)
				del self.pawn_plane._redraw_bind_data_uid
			self.spot_plane.data = self.pawn_plane.data = []
			self.spot_plane.redraw()
			self.pawn_plane.redraw()
			self.spot_plane._redraw_bind_data_uid = self.spot_plane.fbind(
				"data", self.spot_plane._trigger_redraw
			)
			self.pawn_plane._redraw_bind_data_uid = self.pawn_plane.fbind(
				"data", self.pawn_plane._trigger_redraw
			)
			self.character.thing.connect(self.update_from_thing)
			self.character.place.connect(self.update_from_place)
			return
		for spt in spot_data:
			self.spot[spt["name"]] = self.spot_cls(
				board=self, proxy=spt["proxy"]
			)
		self.spot_plane.unbind_uid(
			"data", self.spot_plane._redraw_bind_data_uid
		)
		self.spot_plane.data = spot_data
		self.spot_plane.redraw()
		self.spot_plane._redraw_bind_data_uid = self.spot_plane.fbind(
			"data", self.spot_plane._trigger_redraw
		)
		wide = max(datum["x"] for datum in spot_data) + self.tile_width
		high = max(datum["y"] for datum in spot_data) + self.tile_width
		self.size = self.spot_plane.size = self.pawn_plane.size = wide, high
		pawn_data = list(
			map(
				self.make_pawn,
				filter(
					lambda thing: thing["location"] in self.spot,
					self.character.thing.values(),
				),
			)
		)
		for pwn in pawn_data:
			self.pawn[pwn["name"]] = self.pawn_cls(
				board=self, proxy=pwn["proxy"]
			)
		self.pawn_plane.data = pawn_data
		self.character.thing.connect(self.update_from_thing)
		self.character.place.connect(self.update_from_place)
		Logger.debug(
			f"GridBoard: on_parent end, took {monotonic() - start_ts:,.2f}"
			f" seconds"
		)

	def rm_spot(self, name):
		spot = self.spot.pop(name)
		if spot in self.selection_candidates:
			self.selection_candidates.remove(spot)
		for thing in spot.proxy.contents():
			self.rm_pawn(thing.name)
		self.spot_plane.remove(name)

	def rm_pawn(self, name):
		pwn = self.pawn.pop(name)
		if pwn in self.selection_candidates:
			self.selection_candidates.remove(pwn)
		self.pawn_plane.remove(name)

	def update_from_thing(self, thing, key, value):
		if thing and not (key is None and not value):
			if thing.name not in self.pawn:
				self.add_pawn(thing.name)
			elif key == "location":
				if value in self.spot:
					loc = self.spot[value]
					pwn = self.pawn[thing.name]
					pwn.pos = loc.pos
				elif thing.name in self.pawn:
					self.rm_pawn(thing.name)
			elif key == "_image_paths":
				self.pawn[thing.name].paths = value
		else:
			if thing.name in self.pawn:
				self.rm_pawn(thing.name)

	def update_from_place(self, place, key, value):
		if place and not (key is None and not value):
			if key == "_image_paths":
				self.spot[place.name].paths = value
		else:
			if place.name in self.spot:
				self.rm_spot(place.name)

	def on_touch_down(self, touch):
		self.selection_candidates.extend(
			chain(
				map(
					self.pawn.get,
					self.pawn_plane.iter_collided_keys(*touch.pos),
				),
				map(
					self.spot.get,
					self.spot_plane.iter_collided_keys(*touch.pos),
				),
			)
		)
		return super().on_touch_down(touch)

	def on_touch_up(self, touch):
		touched = {
			candidate
			for candidate in self.selection_candidates
			if candidate.collide_point(*touch.pos)
		}
		if not touched:
			self.selection_candidates = []
			return super().on_touch_up(touch)
		if len(touched) == 1:
			self.selection = touched.pop()
			self.selection_candidates = []
			return super().on_touch_up(touch)
		pawns_touched = {
			node for node in touched if isinstance(node, self.pawn_cls)
		}
		if len(pawns_touched) == 1:
			self.selection = pawns_touched.pop()
			self.selection_candidates = []
			return super().on_touch_up(touch)
		elif pawns_touched:
			# TODO: Repeatedly touching a spot with multiple pawns on it
			# 		should cycle through the pawns, and then finally the spot.
			self.selection = pawns_touched.pop()
			self.selection_candidates = []
			return super().on_touch_up(touch)
		spots_touched = touched - pawns_touched
		if len(spots_touched) == 1:
			self.selection = spots_touched.pop()
			self.selection_candidates = []
			return super().on_touch_up(touch)
		assert not spots_touched, (
			"How do you have overlapping spots on a GridBoard??"
		)
		self.selection_candidates = []
		return super().on_touch_up(touch)


class GridBoardScatterPlane(BoardScatterPlane):
	selection_candidates = ListProperty([])
	selection = ObjectProperty(allownone=True)
	keep_selection = BooleanProperty(False)
	board = ObjectProperty()

	def spot_from_dummy(self, dummy):
		raise NotImplementedError("oop")

	def pawn_from_dummy(self, dummy):
		dummy_center = self.to_local(*dummy.center)
		candidates = list(
			self.board.spot_plane.iter_collided_keys(*dummy_center)
		)
		if not candidates:
			return
		whereat_d = self.board.spot[candidates.pop()]
		half_wide = self.board.tile_width / 2
		half_high = self.board.tile_height / 2
		if candidates:
			whereat_center = whereat_d.x + half_wide, whereat_d.y + half_high
			dist = Vector(*whereat_center).distance(dummy_center)
			while candidates:
				thereat_d = self.board.spot[candidates.pop()]
				thereat_center = (
					thereat_d.x + half_wide,
					thereat_d.y + half_high,
				)
				thereto = Vector(*thereat_center).distance(dummy_center)
				if thereto < dist:
					whereat_d, dist = thereat_d, thereto
		self.board.pawn_plane.add_datum(
			self.board.make_pawn(
				self.board.character.new_thing(
					dummy.name, whereat_d.name, _image_paths=list(dummy.paths)
				)
			)
		)
		dummy.num += 1

	def on_board(self, *args):
		self.clear_widgets()
		self.add_widget(self.board)


class GridBoardView(BoardView):
	pass


store_kv(
	__name__,
	"""
<GridBoard>:
	app: app
	size_hint: None, None
<GridBoardView>:
	plane: boardplane
	GridBoardScatterPlane:
		id: boardplane
		board: root.board
		scale_min: root.scale_min
		scale_max: root.scale_max
		pos: root.pos
		size: root.size
""",
)



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from kivy.clock import Clock, triggered
from kivy.properties import NumericProperty, ObjectProperty

from .util import logwrap


def trigger(func):
	return triggered()(func)


class PawnBehavior:
	"""Mix-in class for things in places represented graphically"""

	loc_name = ObjectProperty()
	default_image_paths = ["atlas://rltiles/base.atlas/unseen"]
	priority = NumericProperty()
	board = ObjectProperty()

	def __init__(self, **kwargs):
		if "thing" in kwargs:
			kwargs["proxy"] = kwargs["thing"]
			del kwargs["thing"]
		if "proxy" in kwargs:
			kwargs["loc_name"] = kwargs["proxy"]["location"]
		super().__init__(**kwargs)
		self.register_event_type("on_drop")

	@logwrap(section="PawnBehavior")
	def on_proxy(self, *args):
		self.loc_name = self.proxy["location"]

	@logwrap(section="PawnBehavior")
	def on_parent(self, *args):
		if not self.parent:
			Clock.schedule_once(self.on_parent, 0)
			return
		self.board = self.parent.board
		self._relocate_binding = self.fbind("loc_name", self._trigger_relocate)
		if self.proxy:
			self._trigger_relocate()

	@logwrap(section="PawnBehavior")
	def finalize(self, initial=True):
		if initial:
			self.loc_name = self.proxy["location"]
			self.priority = self.proxy.get("_priority", 0.0)
		self._push_loc_binding = self.fbind(
			"loc_name", self._trigger_push_location
		)
		super().finalize(initial)

	@logwrap(section="PawnBehavior")
	def unfinalize(self):
		self.unbind_uid("loc_name", self._push_loc_binding)
		super().unfinalize()

	@logwrap(section="PawnBehavior")
	def pull_from_proxy(self, *args):
		super().pull_from_proxy(*args)
		relocate = False
		if self.loc_name != self.proxy["location"]:
			self.unfinalize()
			self.unbind_uid("loc_name", self._relocate_binding)
			self.loc_name = self.proxy["location"]
			self._relocate_binding = self.fbind(
				"loc_name", self._trigger_relocate
			)
			self.finalize(initial=False)
			relocate = True
		if "_priority" in self.proxy:
			self.priority = self.proxy["_priority"]
		if relocate:
			self.relocate()

	@logwrap(section="PawnBehavior")
	def relocate(self, *args):
		if (
			not getattr(self, "_finalized", False)
			or not self.parent
			or not self.proxy
			or not self.proxy.exists
		):
			return
		try:
			location = self._get_location_wid()
		except KeyError:
			return
		if location != self.parent:
			if self.parent:
				self.parent.remove_widget(self)
			location.add_widget(self)

	_trigger_relocate = trigger(relocate)

	@logwrap(section="PawnBehavior")
	def on_priority(self, *args):
		if self.proxy["_priority"] != self.priority:
			self.proxy["_priority"] = self.priority
		self.parent.restack()

	@logwrap(section="PawnBehavior")
	def push_location(self, *args):
		if self.proxy["location"] != self.loc_name:
			self.proxy["location"] = self.loc_name

	_trigger_push_location = trigger(push_location)

	def _get_location_wid(self):
		return self.board.spot[self.loc_name]

	@logwrap(section="PawnBehavior")
	def on_touch_up(self, touch):
		if touch.grab_current is not self:
			return False
		for spot in self.board.spot.values():
			if self.collide_widget(spot) and spot.name != self.loc_name:
				new_spot = spot
				break
		else:
			new_spot = None

		self.dispatch("on_drop", new_spot)
		touch.ungrab(self)
		return True

	@logwrap(section="PawnBehavior")
	def on_drop(self, spot):
		parent = self.parent
		if spot:
			self.loc_name = self.proxy["location"] = spot.name
			parent.remove_widget(self)
			spot.add_widget(self)
		else:
			x, y = getattr(self, "rel_pos", (0, 0))
			self.pos = parent.x + x, parent.y + y



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from functools import partial

from kivy.properties import NumericProperty, OptionProperty
from kivy.uix.boxlayout import BoxLayout
from networkx import grid_2d_graph

from lisien.character import grid_2d_8graph

from .util import logwrap, store_kv


class GridGeneratorDialog(BoxLayout):
	xval = NumericProperty()
	yval = NumericProperty()
	directions = OptionProperty(None, options=[None, 4, 8])

	@partial(logwrap, section="GridGeneratorDialog")
	def generate(self, engine):
		x = int(self.xval)
		y = int(self.yval)
		if x < 1 or y < 1:
			return False
		elif self.directions == 4:
			# instead, we're running just after game init, before the view is open on it, and we'll make a character ourselves
			engine.add_character("physical", grid_2d_graph(x, y))
			return True
		elif self.directions == 8:
			engine.add_character("physical", grid_2d_8graph(x, y))
			return True
		else:
			return False

	@logwrap(section="GridGeneratorDialog")
	def validate(self):
		return self.directions and int(self.xval) and int(self.yval)


store_kv(
	__name__,
	"""
<GridGeneratorDialog>:
	directions: 4 if but4.state == 'down' else 8
	orientation: 'vertical'
	BoxLayout:
		orientation: 'horizontal'
		MenuIntInput:
			id: input_x
			hint_text: str(root.xval) if root.xval else 'x'
			set_value: root.setter('xval')
		Label:
			text: 'x'
			size_hint_x: 0.1
		MenuIntInput:
			id: input_y
			hint_text: str(root.yval) if root.yval else 'y'
			set_value: root.setter('yval')
	BoxLayout:
		ToggleButton:
			id: but4
			group: 'dir'
			text: '4-way'
			state: 'down'
		ToggleButton:
			id: but8
			group: 'dir'
			text: '8-way'""",
)



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from functools import partial

from kivy.app import App
from kivy.clock import Clock
from kivy.properties import ListProperty, ObjectProperty, StringProperty
from kivy.uix.recycleview import RecycleView
from kivy.uix.screenmanager import Screen

from .util import SelectableRecycleBoxLayout, logwrap, store_kv

# TODO: Visual preview
# TODO: Background image chooser


class CharactersRecycleBoxLayout(SelectableRecycleBoxLayout):
	character_name = StringProperty()

	@partial(logwrap, section="CharactersRecycleBoxLayout")
	def apply_selection(self, index, view, is_selected):
		super().apply_selection(index, view, is_selected)
		if is_selected:
			self.character_name = view.text


class CharactersView(RecycleView):
	character_name = StringProperty()

	def __init__(self, **kwargs):
		self.i2name = {}
		self.name2i = {}
		super().__init__(**kwargs)


class CharactersScreen(Screen):
	toggle = ObjectProperty()
	charsview = ObjectProperty()
	character_name = StringProperty()
	wallpaper_path = StringProperty()
	names = ListProperty()
	new_board = ObjectProperty()
	push_character_name = ObjectProperty()

	@property
	def engine(self):
		return App.get_running_app().engine

	@logwrap(section="CharactersScreen")
	def new_character(self, name, *_):
		self.engine.add_character(name)
		self.ids.newname.text = ""
		i = len(self.charsview.data)
		self.charsview.i2name[i] = name
		self.charsview.name2i[name] = i
		self.charsview.data.append({"index": i, "text": name})
		self.names.append(name)
		self.new_board(name)
		self.push_character_name(name)

	@logwrap(section="CharactersScreen")
	def _trigger_new_character(self, name):
		part = partial(self.new_character, name)
		if hasattr(self, "_scheduled_new_character"):
			Clock.unschedule(self._scheduled_new_character)
		self._scheduled_new_character = Clock.schedule_once(part)

	@logwrap(section="CharactersScreen")
	def _munge_names(self, names):
		for i, name in enumerate(names):
			self.charsview.i2name[i] = name
			self.charsview.name2i[name] = i
			yield {"index": i, "text": name}

	@logwrap(section="CharactersScreen")
	def on_names(self, *_):
		app = App.get_running_app()
		if not app.character or not self.charsview:
			Clock.schedule_once(self.on_names, 0)
			return
		self.charsview.data = list(self._munge_names(self.names))
		charname = app.character.name
		for i, name in enumerate(self.names):
			if name == charname:
				self.charsview.children[0].select_node(i)
				return

	@logwrap(section="CharactersScreen")
	def on_charsview(self, *_):
		if not self.push_character_name:
			Clock.schedule_once(self.on_charsview, 0)
			return
		self.charsview.bind(character_name=self.setter("character_name"))
		self.bind(character_name=self.push_character_name)


store_kv(
	__name__,
	"""
#: import resource_find kivy.resources.resource_find
<CharactersView>:
	viewclass: 'RecycleToggleButton'
	character_name: boxl.character_name
	CharactersRecycleBoxLayout:
		id: boxl
		multiselect: False
		default_size: None, dp(56)
		default_size_hint: 1, None
		size_hint_y: None
		height: self.minimum_height
		orientation: 'vertical'
<CharactersScreen>:
	name: 'chars'
	charsview: charsview
	BoxLayout:
		id: chars
		orientation: 'vertical'
		CharactersView:
			id: charsview
			size_hint_y: 0.8
			character_name: root.character_name
		TextInput:
			id: newname
			size_hint_y: 0.1
			hint_text: 'New character name'
			write_tab: False
			multiline: False
		Button:
			text: '+'
			on_release: root._trigger_new_character(newname.text)
			size_hint_y: 0.05
		Button:
			text: 'Close'
			on_release: root.toggle()
			size_hint_y: 0.05
""",
)



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from functools import partial

from kivy.clock import Clock, triggered
from kivy.properties import (
	BooleanProperty,
	ObjectProperty,
	ReferenceListProperty,
)
from kivy.uix.boxlayout import BoxLayout

from lisien.proxy import CharStatProxy

from .graph.arrow import GraphArrowWidget
from .util import dummynum, logwrap, store_kv


def trigger(func):
	return triggered()(func)


class CharMenu(BoxLayout):
	screen = ObjectProperty()
	reciprocal_portal = BooleanProperty(True)
	revarrow = ObjectProperty(None, allownone=True)
	dummyplace = ObjectProperty()
	dummything = ObjectProperty()
	toggle_gridview = ObjectProperty()
	toggle_timestream = ObjectProperty()
	dummies = ReferenceListProperty(dummyplace, dummything)

	@property
	def app(self):
		if not self.screen:
			raise AttributeError("No screen, therefore no app")
		return self.screen.app

	@property
	def engine(self):
		if not self.screen or not self.screen.app:
			raise AttributeError("Can't get engine from screen")
		return self.screen.app.engine

	@logwrap(section="CharMenu")
	def on_screen(self, *_):
		if not (
			self.screen
			and self.screen.boardview
			and self.screen.app
			and "emptyleft" in self.ids
			and "emptyright" in self.ids
		):
			Clock.schedule_once(self.on_screen, 0)
			return
		self.forearrow = GraphArrowWidget(
			board=self.screen.boardview.board,
			origin=self.ids.emptyleft,
			destination=self.ids.emptyright,
		)
		self.ids.portaladdbut.add_widget(self.forearrow)
		self.ids.emptyleft.bind(pos=self.forearrow._trigger_repoint)
		self.ids.emptyright.bind(pos=self.forearrow._trigger_repoint)
		if self.reciprocal_portal:
			assert self.revarrow is None
			self.revarrow = GraphArrowWidget(
				board=self.screen.boardview.board,
				origin=self.ids.emptyright,
				destination=self.ids.emptyleft,
			)
			self.ids.portaladdbut.add_widget(self.revarrow)
			self.ids.emptyleft.bind(pos=self.revarrow._trigger_repoint)
			self.ids.emptyright.bind(pos=self.revarrow._trigger_repoint)
		self.bind(
			reciprocal_portal=self.screen.boardview.setter("reciprocal_portal")
		)

	@logwrap(section="CharMenu")
	def spot_from_dummy(self, dummy):
		if self.screen.boardview.parent != self.screen.mainview:
			return
		if dummy.collide_widget(self):
			return
		name = dummy.name
		self.screen.boardview.spot_from_dummy(dummy)
		graphboard = self.screen.graphboards[self.app.character_name]
		if name not in graphboard.spot:
			graphboard.add_spot(name)
		gridboard = self.screen.gridboards[self.app.character_name]
		if (
			name not in gridboard.spot
			and isinstance(name, tuple)
			and len(name) == 2
		):
			gridboard.add_spot(name)

	@logwrap(section="CharMenu")
	def pawn_from_dummy(self, dummy):
		name = dummy.name
		if not self.screen.mainview.children[0].pawn_from_dummy(dummy):
			return
		graphboard = self.screen.graphboards[self.app.character_name]
		if name not in graphboard.pawn:
			graphboard.add_pawn(name)
		gridboard = self.screen.gridboards[self.app.character_name]
		if (
			name not in gridboard.pawn
			and self.app.character.thing[name]["location"] in gridboard.spot
		):
			gridboard.add_pawn(name)

	@logwrap(section="CharMenu")
	def toggle_chars_screen(self, *_):
		"""Display or hide the list you use to switch between characters."""
		# TODO: update the list of chars
		self.app.chars.toggle()

	@logwrap(section="CharMenu")
	def toggle_rules(self, *_):
		"""Display or hide the view for constructing rules out of cards."""
		if self.app.manager.current != "rules" and not isinstance(
			self.app.selected_proxy, CharStatProxy
		):
			self.app.rules.entity = self.app.selected_proxy
			self.app.rules.rulebook = self.app.selected_proxy.rulebook
		if isinstance(self.app.selected_proxy, CharStatProxy):
			self.app.charrules.character = self.app.selected_proxy
			self.app.charrules.toggle()
		else:
			self.app.rules.toggle()

	@logwrap(section="CharMenu")
	def toggle_funcs_editor(self):
		"""Display or hide the text editing window for functions."""
		self.app.funcs.toggle()

	@logwrap(section="CharMenu")
	def toggle_strings_editor(self):
		self.app.strings.toggle()

	@logwrap(section="CharMenu")
	def toggle_spot_cfg(self):
		"""Show the dialog where you select graphics and a name for a place,
		or hide it if already showing.

		"""
		if self.app.manager.current == "spotcfg":
			dummyplace = self.screendummyplace
			self.ids.placetab.remove_widget(dummyplace)
			dummyplace.clear()
			if self.app.spotcfg.prefix:
				dummyplace.prefix = self.app.spotcfg.prefix
				dummyplace.num = (
					dummynum(self.app.character, dummyplace.prefix) + 1
				)
			if self.app.spotcfg.imgpaths:
				dummyplace.paths = self.app.spotcfg.imgpaths
			else:
				dummyplace.paths = ["atlas://rltiles/floor/floor-stone"]
			dummyplace.center = self.ids.placetab.center
			self.ids.placetab.add_widget(dummyplace)
		else:
			self.app.spotcfg.prefix = self.ids.dummyplace.prefix
		self.app.spotcfg.toggle()

	@logwrap(section="CharMenu")
	def toggle_pawn_cfg(self):
		"""Show or hide the pop-over where you can configure the dummy pawn"""
		if self.app.manager.current == "pawncfg":
			dummything = self.app.dummything
			self.ids.thingtab.remove_widget(dummything)
			dummything.clear()
			if self.app.pawncfg.prefix:
				dummything.prefix = self.app.pawncfg.prefix
				dummything.num = (
					dummynum(self.app.character, dummything.prefix) + 1
				)
			if self.app.pawncfg.imgpaths:
				dummything.paths = self.app.pawncfg.imgpaths
			else:
				dummything.paths = ["atlas://rltiles/base/unseen"]
			self.ids.thingtab.add_widget(dummything)
		else:
			self.app.pawncfg.prefix = self.ids.dummything.prefix
		self.app.pawncfg.toggle()

	@logwrap(section="CharMenu")
	def toggle_reciprocal(self):
		"""Flip my ``reciprocal_portal`` boolean, and draw (or stop drawing)
		an extra arrow on the appropriate button to indicate the
		fact.

		"""
		self.reciprocal_portal = (
			self.screen.boardview.reciprocal_portal
		) = not self.screen.boardview.reciprocal_portal
		if self.screen.boardview.reciprocal_portal:
			assert self.revarrow is None
			self.revarrow = GraphArrowWidget(
				board=self.screen.boardview.board,
				origin=self.ids.emptyright,
				destination=self.ids.emptyleft,
			)
			self.ids.portaladdbut.add_widget(self.revarrow)
			self.ids.emptyright.bind(pos=self.revarrow._trigger_repoint)
			self.ids.emptyleft.bind(pos=self.revarrow._trigger_repoint)
		else:
			if hasattr(self, "revarrow"):
				self.ids.portaladdbut.remove_widget(self.revarrow)
				self.revarrow = None

	@logwrap(section="CharMenu")
	def new_character(self, but):
		name = self.app.chars.ids.newname.text
		try:
			charn = self.app.engine.unpack(name)
		except (TypeError, ValueError):
			charn = name
		self.app.select_character(self.app.engine.new_character(charn))
		self.app.chars.ids.newname.text = ""
		self.app.chars.charsview.adapter.data = list(
			self.engine.character.keys()
		)
		Clock.schedule_once(self.toggle_chars_screen, 0.01)

	@logwrap(section="CharMenu")
	def on_dummyplace(self, *_):
		if not self.dummyplace.paths:
			self.dummyplace.paths = ["atlas://rltiles/floor.atlas/floor-stone"]

	@logwrap(section="CharMenu")
	def on_dummything(self, *_):
		if not self.dummything.paths:
			self.dummything.paths = ["atlas://rltiles/base.atlas/unseen"]

	@trigger
	@logwrap(section="CharMenu")
	def _trigger_deselect(self, *_):
		if hasattr(self.app.selection, "selected"):
			self.app.selection.selected = False
		self.app.selection = None


store_kv(
	__name__,
	"""
<CharMenu>:
	orientation: 'vertical'
	dummyplace: dummyplace
	dummything: dummything
	portaladdbut: portaladdbut
	portaldirbut: portaldirbut
	Button:
		text: 'Delete'
		disabled: app.edit_locked
		on_release: app.delete_selection()
	Button:
		id: timestreambut
		text: 'Timestream'
		disabled: app.edit_locked
		on_release: root.toggle_timestream()
	Button:
		id: gridviewbut
		text: 'Toggle grid'
		on_release: root.toggle_gridview()
	Button:
		text: 'Strings'
		disabled: app.edit_locked
		on_release: root.toggle_strings_editor()
	Button:
		text: 'Python'
		disabled: app.edit_locked
		on_release: root.toggle_funcs_editor()
	Button:
		text: 'Rules'
		disabled: app.edit_locked
		on_release: root.toggle_rules()
	Button:
		text: 'Characters'
		on_release: root.toggle_chars_screen()
	BoxLayout:
		Widget:
			id: placetab
			Dummy:
				id: dummyplace
				center: placetab.center
				prefix: 'place'
				disabled: app.edit_locked
				on_pos_up: root.spot_from_dummy(self)
		Button:
			text: 'cfg'
			disabled: app.edit_locked
			on_release: root.toggle_spot_cfg()
	BoxLayout:
		Widget:
			id: thingtab
			Dummy:
				id: dummything
				center: thingtab.center
				prefix: 'thing'
				disabled: app.edit_locked
				on_pos_up: root.pawn_from_dummy(self)
		Button:
			text: 'cfg'
			disabled: app.edit_locked
			on_release: root.toggle_pawn_cfg()
	BoxLayout:
		orientation: 'vertical'
		ToggleButton:
			id: portaladdbut
			disabled: app.edit_locked
			on_state: root._trigger_deselect()
			Widget:
				id: emptyleft
				center_x: portaladdbut.x + portaladdbut.width / 3
				center_y: portaladdbut.center_y
				size: (0, 0)
			Widget:
				id: emptyright
				center_x: portaladdbut.right - portaladdbut.width / 3
				center_y: portaladdbut.center_y
				size: (0, 0)
		Button:
			id: portaldirbut
			text: 'One-way' if root.reciprocal_portal else 'Two-way'
			disabled: app.edit_locked
			on_release: root.toggle_reciprocal()
	Button:
		text: 'Quit'
		on_release: app.close_game()
""",
)



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from kivy.logger import ConsoleHandler, KivyFormatter, Logger
from kivy.resources import resource_add_path

formatter = KivyFormatter("%(asctime)s [%(levelname)-7s] %(message)s")
for handler in Logger.handlers:
	if not isinstance(handler, ConsoleHandler):
		handler.setFormatter(formatter)

resource_add_path(__path__[0] + "/assets")

__all__ = [
	"graph",
	"grid",
	"app",
	"card",
	"dialog",
	"game",
	"menu",
	"spritebuilder",
	"calendar",
]



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import os
import shutil
import zipfile
from functools import partial

from kivy import Logger
from kivy.app import App
from kivy.clock import Clock, mainthread, triggered
from kivy.properties import ObjectProperty, OptionProperty, StringProperty
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.button import Button
from kivy.uix.dropdown import DropDown
from kivy.uix.filechooser import FileChooserIconView
from kivy.uix.label import Label
from kivy.uix.modalview import ModalView
from kivy.uix.recycleview import RecycleView
from kivy.uix.screenmanager import Screen
from kivy.uix.textinput import TextInput

from .gen import GridGeneratorDialog
from .util import logwrap


class MenuTextInput(TextInput):
	"""Special text input for setting the branch"""

	set_value = ObjectProperty()

	def __init__(self, **kwargs):
		"""Disable multiline, and bind ``on_text_validate`` to ``on_enter``"""
		kwargs["multiline"] = False
		super().__init__(**kwargs)
		self.bind(on_text_validate=self.on_enter)

	@logwrap(section="MenuTextInput")
	def on_enter(self, *_):
		"""Call the setter and blank myself out so that my hint text shows
		up. It will be the same you just entered if everything's
		working.

		"""
		if self.text == "":
			return
		self.set_value(Clock.get_time(), self.text)
		self.text = ""
		self.focus = False

	@logwrap(section="MenuTextInput")
	def on_focus(self, *args):
		"""If I've lost focus, treat it as if the user hit Enter."""
		if not self.focus:
			self.on_enter(*args)

	@logwrap(section="MenuTextInput")
	def on_text_validate(self, *_):
		"""Equivalent to hitting Enter."""
		self.on_enter()


class MenuIntInput(MenuTextInput):
	"""Special text input for setting the turn or tick"""

	@logwrap(section="MenuIntInput")
	def insert_text(self, s, from_undo=False):
		"""Natural numbers only."""
		return super().insert_text(
			"".join(c for c in s if c in "0123456789"), from_undo
		)


class GeneratorButton(Button):
	pass


class WorldStartConfigurator(BoxLayout):
	"""Give options for how to initialize the world state"""

	generator_type = OptionProperty("none", options=["none", "grid"])
	dismiss = ObjectProperty()
	init_board = ObjectProperty()

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.grid_config = GridGeneratorDialog()
		self.generator_dropdown = DropDown()

		def select_txt(btn):
			self.generator_dropdown.select(btn.text)

		for opt in ["None", "Grid"]:
			self.generator_dropdown.add_widget(
				GeneratorButton(text=opt, on_release=select_txt)
			)
		self.generator_dropdown.bind(on_select=self.select_generator_type)

	@logwrap(section="WorldStartConfigurator")
	def select_generator_type(self, instance, value):
		self.ids.drop.text = value
		if value == "none":
			self.ids.controls.clear_widgets()
			self.generator_type = "none"
		elif value == "Grid":
			self.ids.controls.clear_widgets()
			self.ids.controls.add_widget(self.grid_config)
			self.grid_config.size = self.ids.controls.size
			self.grid_config.pos = self.ids.controls.pos
			self.generator_type = "grid"


class GamePickerModal(ModalView):
	headline = StringProperty()

	@logwrap(section="GamePickerModal")
	def _decompress_and_start(self, game_file_path, game, *_):
		app = App.get_running_app()
		game_dir = str(os.path.join(app.prefix, game))
		if os.path.exists(game_dir):
			# Likely left over from a failed run of Elide
			shutil.rmtree(game_dir)
		shutil.unpack_archive(game_file_path, game_dir)
		Clock.schedule_once(partial(app.start_game, name=game), 0.001)
		self.dismiss(force=True)


class GameExporterModal(GamePickerModal):
	@triggered()
	@logwrap(section="GameExporterModal")
	def pick(self, game, *_):
		app = App.get_running_app()
		app.copy_to_shared_storage(
			os.path.join(app.games_dir, game + ".zip"), "application/zip"
		)
		self.dismiss()

	@logwrap(section="GameExporterModal")
	def regen(self):
		if "game_list" not in self.ids:
			return
		self.ids.game_list.regen()


class GameImporterModal(GamePickerModal):
	@triggered()
	@logwrap(section="GameImporterModal")
	def pick(self, selection, *_):
		if not selection:
			return
		if len(selection) > 1:
			raise RuntimeError(
				"That file picker is supposed to be single select"
			)
		uri = selection[0]
		if isinstance(uri, str):
			uri_s = uri
		else:
			uri_s = uri.toString()
		if os.path.isdir(uri_s):
			return
		try:
			game_name = os.path.basename(uri).removesuffix(".zip")
			self._decompress_and_start(uri, game_name)
		except (
			NotADirectoryError,
			FileNotFoundError,
			FileExistsError,
			zipfile.error,
		) as err:
			Logger.error(repr(err))
			modal = ModalView()
			error_box = BoxLayout(orientation="vertical")
			error_box.add_widget(Label(text=repr(err), font_size=80))
			error_box.add_widget(Button(text="OK", on_release=modal.dismiss))
			modal.add_widget(error_box)
			modal.open()

	@logwrap(section="GameImporterModal")
	def on_pre_open(self, *_):
		try:
			from android.storage import primary_external_storage_path

			Logger.error(
				"GameImporterModal: running on Android, where it won't work"
			)
			path = primary_external_storage_path()
			self._android = True
		except ImportError:
			path = os.getcwd()
			self._android = False
		if not hasattr(self, "_file_chooser"):
			self._file_chooser = FileChooserIconView(path=path)
			self.ids.chooser_goes_here.add_widget(self._file_chooser)


class GameLoaderModal(GamePickerModal):
	@triggered()
	@logwrap(section="GameLoaderModal")
	def pick(self, game, *_):
		app = App.get_running_app()
		if os.path.isfile(app.games_dir):
			raise RuntimeError(
				"You put a file where I want to keep the games directory",
				app.games_dir,
			)
		if not os.path.exists(app.games_dir):
			os.makedirs(app.games_dir)
		if game + ".zip" in os.listdir(app.games_dir):
			game_file_path = str(os.path.join(app.games_dir, game + ".zip"))
			if not zipfile.is_zipfile(game_file_path):
				raise RuntimeError("Game format invalid", game_file_path)
		else:
			raise RuntimeError("Invalid game name", game)
		self.clear_widgets()
		self.add_widget(Label(text="Please wait...", font_size=80))
		Clock.schedule_once(
			partial(self._decompress_and_start, game_file_path, game), 0.05
		)


class GameList(RecycleView):
	picker = ObjectProperty()
	path = StringProperty()

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self._trigger_regen = Clock.create_trigger(self.regen)
		self.bind(picker=self._trigger_regen, path=self._trigger_regen)

	@logwrap(section="GameList")
	def regen(self, *_):
		if not self.picker:
			Logger.debug("GameList: awaiting picker")
			Clock.schedule_once(self.on_path, 0)
			return
		if not os.path.isdir(self.path):
			return
		self.data = [
			{
				"text": game.removesuffix(".zip"),
				"on_release": partial(
					self.picker.pick, game.removesuffix(".zip")
				),
			}
			for game in filter(
				lambda game: game.endswith(".zip")
				and not game.startswith("."),
				os.listdir(self.path or "."),
			)
		]
		Logger.debug(f"GameList: generated {len(self.data)} entries")


class NewGameModal(ModalView):
	@triggered()
	@logwrap(section="NewGameModal")
	def validate_and_start(self, *_):
		game_name = self.ids.game_name.text
		self.ids.game_name.text = ""
		if not game_name:
			self.ids.game_name.hint_text = "Must be nonempty"
			return
		app = App.get_running_app()
		if os.path.isdir(app.games_dir):
			games = [
				fn.removesuffix(".zip") for fn in os.listdir(app.games_dir)
			]
		else:
			os.makedirs(app.games_dir)
			games = []
		if game_name in games:
			self.ids.game_name.hint_text = "Name already taken"
			return
		game_archive_path = os.path.join(app.games_dir, game_name + ".zip")
		game_dir_path = os.path.join(app.prefix, game_name)
		can_start = False
		try:
			zipfile.ZipFile(game_archive_path, "w").close()
			os.makedirs(game_dir_path)
			can_start = True
		except Exception as ex:
			self.ids.game_name.hint_text = repr(ex)
		finally:
			if os.path.isfile(game_archive_path):
				os.remove(game_archive_path)
		if can_start and (
			self.ids.worldstart.generator_type.lower() == "none"
			or self.ids.worldstart.grid_config.validate()
		):
			self.clear_widgets()
			self.add_widget(Label(text="Please wait...", font_size=80))
			self.canvas.ask_update()
			if os.path.exists(app.prefix) and any(
				fn not in {".", ".."} for fn in os.listdir(app.prefix)
			):
				app.close_game()
			Clock.schedule_once(partial(self._really_start, game_name), 0.05)

	@logwrap(section="NewGameModal")
	def _really_start(self, game_name, *_):
		app = App.get_running_app()
		worldstart = self.ids.worldstart
		if worldstart.generator_type == "grid":
			app.start_game(
				name=game_name,
				cb=lambda: worldstart.grid_config.generate(app.engine),
			)
		else:
			app.start_game(
				name=game_name, cb=lambda: app.engine.add_character("physical")
			)
		app.select_character(app.engine.character["physical"])
		self.dismiss()


def trigger(func: callable) -> callable:
	return triggered()(func)


class MainMenuScreen(Screen):
	toggle = ObjectProperty()

	@trigger
	@logwrap(section="MainMenuScreen")
	def new_game(self, *_):
		if not hasattr(self, "_popover_new_game"):
			self._popover_new_game = NewGameModal()
		self._popover_new_game.open()

	@trigger
	@logwrap(section="MainMenuScreen")
	def load_game(self, *_):
		if not hasattr(self, "_popover_load_game"):
			self._popover_load_game = GameLoaderModal(
				headline="Pick game to load"
			)
		self._popover_load_game.open()

	@trigger
	@logwrap(section="MainMenuScreen")
	def import_game(self, *_):
		try:
			import android
			from androidstorage4kivy import Chooser, SharedStorage

			Logger.debug("Using Android system file chooser")

			if not hasattr(self, "_system_file_chooser"):
				self._ss = SharedStorage()
				self._system_file_chooser = Chooser(
					self._copy_from_shared_and_start_game
				)
			self._system_file_chooser.choose_content("application/zip")
		except ImportError as err:
			Logger.debug(repr(err))
			Logger.debug("Using Kivy file chooser")
			if not hasattr(self, "_popover_import_game"):
				self._popover_import_game = GameImporterModal(
					headline="Pick zipped game to import"
				)
			self._popover_import_game.open()

	@mainthread
	@logwrap(section="MainMenuScreen")
	def _copy_from_shared_and_start_game(self, files):
		game_file_path = self._ss.copy_from_shared(files[0])
		if not game_file_path.endswith(".zip"):
			return
		game = str(os.path.basename(game_file_path).removesuffix(".zip"))
		app = App.get_running_app()
		self._please_wait = ModalView()
		self._please_wait.add_widget(
			Label(text="Please wait...", font_size=80)
		)
		self._please_wait.open()
		game_dir = str(os.path.join(app.prefix, game))
		if os.path.exists(game_dir):
			# Likely left over from a failed run of Elide
			shutil.rmtree(game_dir)
		Clock.schedule_once(
			partial(self._unpack_and_open, game_file_path, game_dir, game),
			0.05,
		)

	def _unpack_and_open(self, game_file_path, game_dir, game, *_):
		app = App.get_running_app()
		shutil.unpack_archive(game_file_path, game_dir)
		app.start_game(name=game, cb=self._please_wait.dismiss)

	@trigger
	@logwrap(section="MainMenuScreen")
	def export_game(self, *_):
		if not hasattr(self, "_popover_export_game"):
			self._popover_export_game = GameExporterModal(
				headline="Pick game to export"
			)
		self._popover_export_game.regen()
		self._popover_export_game.open()

	@trigger
	@logwrap(section="MainMenuScreen")
	def invalidate_popovers(self, *_):
		if hasattr(self, "_popover_new_game"):
			del self._popover_new_game
		if hasattr(self, "_popover_load_game"):
			del self._popover_load_game
		if hasattr(self, "_popover_export_game"):
			del self._popover_export_game



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import json
import os
import shutil

from kivy.clock import Clock, triggered
from kivy.core.image import Image
from kivy.logger import Logger
from kivy.properties import (
	ListProperty,
	NumericProperty,
	ObjectProperty,
	StringProperty,
)
from kivy.resources import resource_find
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.button import Button
from kivy.uix.filechooser import FileChooserIconView
from kivy.uix.label import Label
from kivy.uix.modalview import ModalView
from kivy.uix.screenmanager import Screen
from kivy.uix.scrollview import ScrollView
from sqlalchemy import and_, bindparam, column

from .kivygarden.texturestack import ImageStack
from .pallet import Pallet, PalletBox
from .util import logwrap, store_kv


def trigger(func):
	return triggered()(func)


class SpriteSelector(BoxLayout):
	prefix = StringProperty()
	pallets = ListProperty()
	imgpaths = ListProperty([])
	default_imgpaths = ListProperty()
	preview = ObjectProperty()

	@logwrap(section="SpriteSelector")
	def on_prefix(self, *_):
		if "textbox" not in self.ids:
			Clock.schedule_once(self.on_prefix, 0)
			return
		self.ids.textbox.text = self.prefix

	@logwrap(section="SpriteSelector")
	def on_imgpaths(self, *_):
		if not self.preview:
			Logger.debug("SpriteSelector: no preview")
			Clock.schedule_once(self.on_imgpaths, 0)
			return
		if hasattr(self, "_imgstack"):
			self._imgstack.paths = self.imgpaths
		else:
			self._imgstack = ImageStack(paths=self.imgpaths)
			self._imgstack.bind(
				pos=self._position_imgstack, size=self._position_imgstack
			)
			self.preview.add_widget(self._imgstack)

	@trigger
	@logwrap(section="SpriteSelector")
	def _position_imgstack(self, *_):
		self._imgstack.x = self.preview.center_x - self._imgstack.height / 2
		self._imgstack.y = self.preview.center_y - self._imgstack.width / 2

	@logwrap(section="SpriteSelector")
	def on_pallets(self, *_):
		for pallet in self.pallets:
			pallet.fbind("selection", self._upd_imgpaths)

	@logwrap(section="SpriteSelector")
	def _upd_imgpaths(self, *_):
		imgpaths = []
		for pallet in self.pallets:
			if pallet.selection:
				for selected in pallet.selection:
					imgpaths.append(
						"atlas://{}/{}".format(pallet.filename, selected.text)
					)
		self.imgpaths = imgpaths if imgpaths else self.default_imgpaths


class SpriteBuilder(ScrollView):
	prefix = StringProperty()
	imgpaths = ListProperty()
	default_imgpaths = ListProperty()
	data = ListProperty()
	labels = ListProperty()
	pallets = ListProperty()

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.bind(data=self._trigger_update)

	@logwrap(section="SpriteBuilder")
	def update(self, *args):
		if self.data is None:
			return
		if not self.canvas:
			Clock.schedule_once(self.update, 0)
			return
		if not hasattr(self, "_palbox"):
			self._palbox = PalletBox(orientation="vertical", size_hint_y=None)
			self.add_widget(self._palbox)
		else:
			self._palbox.clear_widgets()
		if hasattr(self._palbox, "_bound_width"):
			for uid in self._palbox._bound_width:
				self._palbox.unbind_uid("width", uid)
			del self._palbox._bound_width
		self.labels = []
		for pallet in self.pallets:
			if hasattr(pallet, "_bound_minimum_height"):
				pallet.unbind_uid(
					"minimum_height", pallet._bound_minimum_height
				)
				del pallet._bound_minimum_height
			if hasattr(pallet, "_bound_height"):
				pallet.unbind_uid("height", pallet._bound_height)
				del pallet._bound_height
		self.pallets = []
		for text, filename in self.data:
			label = Label(text=text, size_hint=(None, None), halign="center")
			label.texture_update()
			label.height = label.texture.height
			label.width = self._palbox.width
			pallet = Pallet(filename=filename, size_hint=(None, None))
			pallet.width = self._palbox.width
			self._palbox._bound_width = [
				self._palbox.fbind("width", label.setter("width")),
				self._palbox.fbind("width", pallet.setter("width")),
			]
			pallet.height = pallet.minimum_height
			pallet._bound_minimum_height = (
				pallet.fbind("minimum_height", pallet.setter("height")),
			)
			pallet._bound_height = pallet.fbind(
				"height", self._trigger_reheight
			)
			self.labels.append(label)
			self.pallets.append(pallet)
		n = len(self.labels)
		assert n == len(self.pallets)
		for i in range(0, n):
			self._palbox.add_widget(self.labels[i])
			self._palbox.add_widget(self.pallets[i])

	_trigger_update = trigger(update)

	@logwrap(section="SpriteBuilder")
	def reheight(self, *args):
		self._palbox.height = sum(
			wid.height for wid in self.labels + self.pallets
		)

	_trigger_reheight = trigger(reheight)


class SpriteDialog(BoxLayout):
	toggle = ObjectProperty()
	prefix = StringProperty()
	imgpaths = ListProperty()
	custom_imgs_header = StringProperty()
	custom_imgs_dir = StringProperty()
	default_imgpaths = ListProperty()
	data = ListProperty()
	pallet_box_height = NumericProperty()

	@logwrap(section="SpriteDialog")
	def pressed(self):
		self.prefix = self.ids.selector.prefix
		self.imgpaths = self.ids.selector.imgpaths
		self.toggle()

	@trigger
	@logwrap(section="SpriteDialog")
	def _choose_graphic_to_import(self, *_):
		try:
			from android.storage import primary_external_storage_path

			path = primary_external_storage_path()
			self._android = True
		except ImportError:
			path = os.getcwd()
			self._android = False
		if hasattr(self, "_graphic_modal"):
			self._file_chooser.path = path
		else:
			self._graphic_modal = ModalView()
			graphic_modal_layout = BoxLayout(orientation="vertical")
			graphic_modal_layout.add_widget(
				Label(text="Pick an image to import", size_hint_y=None)
			)
			self._file_chooser = FileChooserIconView(path=path)
			graphic_modal_layout.add_widget(self._file_chooser)
			cancel = Button(
				text="Cancel", on_release=self._graphic_modal.dismiss
			)
			ok = Button(text="Import", on_release=self._import_graphic)
			buttons_layout = BoxLayout(
				orientation="horizontal", size_hint_y=0.1
			)
			buttons_layout.add_widget(cancel)
			buttons_layout.add_widget(ok)
			graphic_modal_layout.add_widget(buttons_layout)
			self._graphic_modal.add_widget(graphic_modal_layout)
		self._graphic_modal.open()

	@logwrap(section="SpriteDialog")
	def _copy_from_shared(self, src: str):
		from android import autoclass, cast, mActivity
		from android.storage import (
			app_storage_path,
			primary_external_storage_path,
		)

		dst = os.path.join(app_storage_path(), os.path.basename(src))
		MediaStoreMediaColumns = autoclass(
			"android.provider.MediaStore$MediaColumns"
		)
		mtm = autoclass("android.webkit.MimeTypeMap").getSingleton()
		mime_type = mtm.getMimeTypeFromExtension(src[src.rindex(".") + 1 :])
		if not mime_type.lower().startswith("image"):
			raise ValueError("Not an image")
		root_uri = autoclass(
			"android.provider.MediaStore$Files"
		).getContentUri("external")
		context = mActivity.getApplicationContext()
		select_stmt = and_(
			column(MediaStoreMediaColumns.DISPLAY_NAME) == bindparam("a"),
			column(MediaStoreMediaColumns.RELATIVE_PATH) == bindparam("b"),
		)
		select_stmt.stringify_dialect = "sqlite"
		select_s = str(select_stmt)
		Logger.debug(
			f"SpriteDialog: looking for URI using the query: {select_s}"
		)
		args = [
			os.path.basename(src),
			os.path.dirname(src)
			.replace(primary_external_storage_path(), "")
			.strip("/")
			+ "/",
		]
		Logger.debug(f"SpriteDialog: with the aruments: {args}")
		cursor = context.getContentResolver().query(
			root_uri,
			None,
			select_s,
			args,
			None,
		)
		if not cursor:
			raise FileNotFoundError(src)
		while cursor.moveToNext():
			idx = cursor.getColumnIndex(MediaStoreMediaColumns.DISPLAY_NAME)
			file_name = cursor.getString(idx)
			Logger.debug(f"SpriteDialog: file #{idx}. {file_name}")
			if file_name == os.path.basename(src):
				id_ = cursor.getLong(
					cursor.getColumnIndex(MediaStoreMediaColumns._ID)
				)
				uri = autoclass("android.content.ContentUris").withAppendedId(
					root_uri, id_
				)
				break
		else:
			cursor.close()
			raise FileNotFoundError(src)
		if uri.getScheme().lower() == "file":
			shutil.copyfile(src, dst)
			return dst
		context = mActivity.getApplicationContext()
		cursor = context.getContentResolver().query(
			uri, None, None, None, None
		)
		if not cursor:
			raise FileNotFoundError(src)
		if os.path.exists(dst):
			os.remove(dst)
		FileOutputStream = autoclass("java.io.FileOutputStream")
		FileUtils = autoclass("android.os.FileUtils")
		resolver = context.getContentResolver()
		reader = resolver.openInputStream(uri)
		writer = FileOutputStream(dst)
		FileUtils.copy(reader, writer)
		reader.close()
		writer.close()
		cursor.close()
		return dst

	@trigger
	@logwrap(section="SpriteBuilder")
	def _import_graphic(self, *_):
		if not self._file_chooser.selection:
			# should display an error message briefly
			return
		file_path = os.path.join(
			self._file_chooser.path, self._file_chooser.selection[0]
		)
		if self._android:
			if not hasattr(self, "_storage"):
				from android.permissions import Permission, request_permissions

				request_permissions([Permission.READ_MEDIA_IMAGES])
			to_import: str = self._copy_from_shared(file_path)
		else:
			to_import: str = os.path.abspath(file_path)
		# We'll need to load the image before we put it in an atlas.
		# Mainly to get its size, but maybe convert from JPG or BMP or w/e.
		# Kivy seems happiest to work with PNG files.
		# Best to do it first, so that if it fails,
		# no files or dirs are made.
		img = Image.load(to_import)
		if not os.path.exists(self.custom_imgs_dir):
			os.makedirs(self.custom_imgs_dir)
		atlas_fn = os.path.join(self.custom_imgs_dir, "custom.atlas")
		if os.path.exists(atlas_fn):
			with open(atlas_fn, "rt") as inf:
				atlas = json.load(inf)
		else:
			atlas = {}
		dest_fn = f"custom-{len(atlas)}.png"
		dest_path = os.path.join(self.custom_imgs_dir, dest_fn)
		img.save(dest_path)
		atlas[dest_fn] = {
			os.path.basename(to_import): [0, 0, img.width, img.height]
		}
		with open(atlas_fn, "wt") as outf:
			json.dump(atlas, outf)
		if not resource_find(atlas_fn):
			raise FileNotFoundError(
				"The created atlas was not where we saved it", atlas_fn
			)
		self.data = [[self.custom_imgs_header, atlas_fn]] + self.data
		self._graphic_modal.dismiss()


class PawnConfigDialog(SpriteDialog):
	pass


class SpotConfigDialog(SpriteDialog):
	pass


class PawnConfigScreen(Screen):
	toggle = ObjectProperty()
	data = ListProperty()
	custom_imgs_header = "Custom pawns"
	imgpaths = ListProperty()


class SpotConfigScreen(Screen):
	toggle = ObjectProperty()
	data = ListProperty()
	custom_imgs_header = "Custom spots"
	imgpaths = ListProperty()


store_kv(
	__name__,
	"""
<SpriteDialog>:
	orientation: 'vertical'
	SpriteBuilder:
		id: builder
		prefix: root.prefix
		default_imgpaths: root.default_imgpaths
		imgpaths: root.imgpaths
		data: root.data
	SpriteSelector:
		id: selector
		textbox: textbox
		size_hint_y: 0.1
		prefix: root.prefix
		default_imgpaths: root.default_imgpaths
		imgpaths: root.imgpaths
		pallets: builder.pallets
		preview: preview
		TextInput:
			id: textbox
			multiline: False
			write_tab: False
			hint_text: 'Enter name prefix'
		Widget:
			id: preview
			canvas:
				Color:
					rgba: 1, 1, 1, 1
		Button:
			text: 'Import'
			on_release: root._choose_graphic_to_import()
		Button:
			text: 'OK'
			on_release: root.pressed()
<PawnConfigScreen>:
	name: 'pawncfg'
	imgpaths: dialog.imgpaths
	PawnConfigDialog:
		id: dialog
		toggle: root.toggle
		default_imgpaths: ['atlas://rltiles/base/unseen']
		custom_imgs_header: root.custom_imgs_header
		custom_imgs_dir: app.prefix + '/custom_pawn_imgs'
		data: root.data
<SpotConfigScreen>:
	name: 'spotcfg'
	imgpaths: dialog.imgpaths
	SpotConfigDialog:
		id: dialog
		toggle: root.toggle
		default_imgpaths: ['atlas://rltiles/floor/floor-stone']
		custom_imgs_header: root.custom_imgs_header
		custom_imgs_dir: app.prefix + '/custom_spot_imgs'
		data: root.data
""",
)



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from functools import partial
from threading import Thread

from kivy.app import App
from kivy.clock import Clock, triggered
from kivy.factory import Factory
from kivy.logger import Logger
from kivy.properties import (
	BooleanProperty,
	DictProperty,
	NumericProperty,
	ObjectProperty,
	StringProperty,
)
from kivy.uix.screenmanager import NoTransition, Screen, ScreenManager

import lisien.proxy

from .dialog import DialogLayout
from .graph.board import GraphBoard, GraphBoardView
from .grid.board import GridBoard, GridBoardView
from .util import logwrap

Factory.register("GraphBoard", GraphBoard)
Factory.register("GridBoard", GridBoard)
Factory.register("GraphBoardView", GraphBoardView)
Factory.register("GridBoardView", GridBoardView)
Factory.register("DialogLayout", DialogLayout)


class GameScreen(Screen):
	switch_screen = ObjectProperty()
	"""Method to set the ``screen`` attribute of the main :class:`kivy.uix.screenmanager.ScreenManager`"""
	disabled = BooleanProperty(False)
	"""If you bind your widgets' ``disabled`` to this, they will be disabled when a game command is in mid-execution"""

	@property
	def app(self):
		return App.get_running_app()

	@property
	def engine(self):
		return App.get_running_app().engine

	@wraplog_GameScreen
	def disable_input(self, cb=None):
		"""Set ``self.disabled`` to ``True``, then call ``cb`` if provided

		:param cb: callback function for after disabling
		:return: ``None``

		"""
		self.disabled = True
		if cb:
			cb()

	@wraplog_GameScreen
	def enable_input(self, cb=None):
		"""Call ``cb`` if provided, then set ``self.disabled`` to ``False``

		:param cb: callback function for before enabling
		:return: ``None``

		"""
		if cb:
			cb()
		self.disabled = False

	@wraplog_GameScreen
	def wait_travel(self, character, thing, dest, cb=None):
		"""Schedule a thing to travel someplace, then wait for it to finish.

		:param character: name of the character
		:param thing: name of the thing that will travel
		:param dest: name of the place it will travel to
		:param cb: callback function for when it's done, optional
		:return: ``None``

		"""
		self.disable_input()
		self.app.wait_travel(
			character, thing, dest, cb=partial(self.enable_input, cb)
		)

	@wraplog_GameScreen
	def wait_turns(self, turns, cb=None):
		"""Call ``self.app.engine.next_turn()`` ``n`` times, waiting ``self.app.turn_length`` in between

		Disables input for the duration.

		:param turns: number of turns to wait
		:param cb: function to call when done waiting, optional
		:return: ``None``

		"""
		self.disable_input()
		self.app.wait_turns(turns, cb=partial(self.enable_input, cb))

	@wraplog_GameScreen
	def wait_command(self, start_func, turns=1, end_func=None):
		"""Call ``start_func``, wait ``turns``, and then call ``end_func`` if provided

		Disables input for the duration.

		:param start_func: function to call just after disabling input
		:param turns: number of turns to wait
		:param end_func: function to call just before re-enabling input
		:return: ``None``

		"""
		self.disable_input()
		start_func()
		self.app.wait_turns(turns, cb=partial(self.enable_input, end_func))

	@wraplog_GameScreen
	def wait_travel_command(
		self,
		character,
		thing,
		dest,
		start_func,
		turns=1,
		end_func=lambda: None,
	):
		"""Schedule a thing to travel someplace and do something, then wait for it to finish.

		Input will be disabled for the duration.

		:param character: name of the character
		:param thing: name of the thing
		:param dest: name of the destination (a place)
		:param start_func: function to call when the thing gets to dest
		:param turns: number of turns to wait after start_func before re-enabling input
		:param end_func: optional. Function to call after waiting ``turns`` after start_func
		:return: ``None``

		"""
		self.disable_input()
		self.app.wait_travel_command(
			character,
			thing,
			dest,
			start_func,
			turns,
			partial(self.enable_input, end_func),
		)


wraplog_GameApp = partial(logwrap, section="GameApp")


class GameApp(App):
	modules = []
	turn_length = NumericProperty(0.5)
	branch = StringProperty("trunk")
	turn = NumericProperty(0)
	tick = NumericProperty(0)
	prefix = StringProperty(".")
	selection = ObjectProperty(allownone=True)
	engine_kwargs = DictProperty({})

	@logwrap(section="GameApp")
	def wait_turns(self, turns, *, cb=None):
		"""Call ``self.engine.next_turn()`` ``turns`` times, waiting ``self.turn_length`` in between

		If provided, call ``cb`` when done.

		:param turns: number of turns to wait
		:param dt: unused, just satisfies the clock
		:param cb: callback function to call when done, optional
		:return: ``None``

		"""
		if hasattr(self, "_next_turn_thread"):
			Clock.schedule_once(partial(self.wait_turns, turns, cb=cb), 0)
			return
		if turns == 0:
			if cb:
				cb()
			return
		self.next_turn()
		turns -= 1
		Clock.schedule_once(
			partial(self.wait_turns, turns, cb=cb), self.turn_length
		)

	@logwrap(section="GameApp")
	def wait_travel(self, character, thing, dest, cb=None):
		"""Schedule a thing to travel someplace, then wait for it to finish, and call ``cb`` if provided

		:param character: name of the character
		:param thing: name of the thing
		:param dest: name of the destination (a place)
		:param cb: function to be called when I'm done
		:return: ``None``

		"""
		self.wait_turns(
			self.engine.character[character].thing[thing].travel_to(dest),
			cb=cb,
		)

	@logwrap(section="GameApp")
	def wait_command(self, start_func, turns=1, end_func=None):
		"""Call ``start_func``, and wait to call ``end_func`` after simulating ``turns`` (default 1)

		:param start_func: function to call before waiting
		:param turns: number of turns to wait
		:param end_func: function to call after waiting
		:return: ``None``

		"""
		start_func()
		self.wait_turns(turns, cb=end_func)

	@logwrap(section="GameApp")
	def wait_travel_command(
		self, character, thing, dest, start_func, turns=1, end_func=None
	):
		"""Schedule a thing to travel someplace and do something, then wait for it to finish.

		:param character: name of the character
		:param thing: name of the thing
		:param dest: name of the destination (a place)
		:param start_func: function to call when the thing gets to dest
		:param turns: number of turns to wait after start_func before re-enabling input
		:param end_func: optional. Function to call after waiting ``turns`` after start_func
		:return: ``None``
		"""
		self.wait_travel(
			character,
			thing,
			dest,
			cb=partial(self.wait_command, start_func, turns, end_func),
		)

	@logwrap(section="GameApp")
	def _pull_time(self, *_, then, now):
		self.branch, self.turn, self.tick = now

	@logwrap(section="GameApp")
	def build(self):
		self.procman = lisien.proxy.EngineProcessManager()
		self.engine = self.procman.start(
			self.prefix,
			logger=Logger,
			loglevel=getattr(self, "loglevel", "debug"),
			install_modules=self.modules,
			**self.engine_kwargs,
		)
		self.branch, self.turn, self.tick = self.engine._btt()
		self.engine.time.connect(self._pull_time, weak=False)
		self.screen_manager = ScreenManager(transition=NoTransition())
		if hasattr(self, "inspector"):
			from kivy.core.window import Window
			from kivy.modules import inspector

			inspector.create_inspector(Window, self.screen_manager)
		return self.screen_manager

	@logwrap(section="GameApp")
	def on_pause(self):
		"""Sync the database with the current state of the game."""
		self.engine.commit()
		self.config.write()

	@logwrap(section="GameApp")
	def on_stop(self, *_):
		"""Sync the database, wrap up the game, and halt."""
		self.procman.shutdown()
		self.config.write()

	@logwrap(section="GameApp")
	def _del_next_turn_thread(self, *_, **__):
		del self._next_turn_thread

	@logwrap(section="GameApp")
	def next_turn(self, *_):
		"""Smoothly advance to the next turn in the simulation

		This uses a subthread to wait for lisien to finish simulating
		the turn and report the changes. The interface will remain responsive.

		If you're wiring up the interface, consider binding user
		input to `trigger_next_turn` instead, so that the user doesn't
		mistakenly go two or three turns into the future.

		"""
		if hasattr(self, "_next_turn_thread"):
			return
		self._next_turn_thread = Thread(
			target=self.engine.next_turn,
			kwargs={"cb": self._del_next_turn_thread},
		)
		self._next_turn_thread.start()

	trigger_next_turn = triggered(next_turn)



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from functools import partial

from kivy.graphics.transformation import Matrix
from kivy.uix.scatter import ScatterPlane

from .util import logwrap

wraplog_bsp = partial(logwrap, section="BoardScatterPlane")


class BoardScatterPlane(ScatterPlane):
	@wraplog_bsp
	def on_touch_down(self, touch):
		if touch.is_mouse_scrolling:
			scale = self.scale + (
				0.05 if touch.button == "scrolldown" else -0.05
			)
			if (self.scale_min and scale < self.scale_min) or (
				self.scale_max and scale > self.scale_max
			):
				return
			rescale = scale * 1.0 / self.scale
			self.apply_transform(
				Matrix().scale(rescale, rescale, rescale),
				post_multiply=True,
				anchor=self.to_local(*touch.pos),
			)
			return self.dispatch("on_transform_with_touch", touch)
		return super().on_touch_down(touch)

	@wraplog_bsp
	def apply_transform(self, trans, post_multiply=False, anchor=(0, 0)):
		super().apply_transform(
			trans, post_multiply=post_multiply, anchor=anchor
		)
		self._last_transform = trans, post_multiply, anchor

	@wraplog_bsp
	def on_transform_with_touch(self, touch):
		x, y = self.pos
		w = self.board.width * self.scale
		h = self.board.height * self.scale
		if hasattr(self, "_last_transform") and (
			w < self.parent.width or h < self.parent.height
		):
			trans, post_multiply, anchor = self._last_transform
			super().apply_transform(trans.inverse(), post_multiply, anchor)
			return
		if x > self.parent.x:
			self.x = self.parent.x
		if y > self.parent.y:
			self.y = self.parent.y
		if x + w < self.parent.right:
			self.x = self.parent.right - w
		if y + h < self.parent.top:
			self.y = self.parent.top - h



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Generic dialog boxes and menus, for in front of a Board"""

from functools import partial

from kivy.app import App
from kivy.clock import Clock, mainthread
from kivy.core.text import DEFAULT_FONT
from kivy.logger import Logger
from kivy.properties import (
	DictProperty,
	ListProperty,
	NumericProperty,
	ObjectProperty,
	StringProperty,
	VariableListProperty,
)
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.button import Button
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.scrollview import ScrollView
from kivy.uix.widget import Widget

from .util import logwrap, store_kv


class Box(Widget):
	padding = VariableListProperty(6)
	border = VariableListProperty(4)
	font_size = StringProperty("15sp")
	font_name = StringProperty(DEFAULT_FONT)
	background = StringProperty()
	background_color = VariableListProperty([1, 1, 1, 1])
	foreground_color = VariableListProperty([0, 0, 0, 1])


class ScrollableLabel(ScrollView):
	font_size = StringProperty("15sp")
	font_name = StringProperty(DEFAULT_FONT)
	color = VariableListProperty([0, 0, 0, 1])
	line_spacing = NumericProperty(0)
	text = StringProperty()


class MessageBox(Box):
	"""Looks like a TextInput but doesn't accept any input.

	Does support styled text with BBcode.

	"""

	line_spacing = NumericProperty(0)
	text = StringProperty()


class DialogMenu(Box):
	"""Some buttons that make the game do things.

	Set ``options`` to a list of pairs of ``(text, function)`` and the
	menu will be populated with buttons that say ``text`` that call
	``function`` when pressed.

	"""

	options = ListProperty()
	"""List of pairs of (button_text, callable)"""

	def _set_sv_size(self, *_):
		self._sv.width = self.width - self.padding[0] - self.padding[2]
		self._sv.height = self.height - self.padding[1] - self.padding[3]

	def _set_sv_pos(self, *_):
		self._sv.x = self.x + self.padding[0]
		self._sv.y = self.y + self.padding[3]

	@mainthread
	@partial(logwrap, section="DialogMenu")
	def on_options(self, *_):
		if not hasattr(self, "_sv"):
			self._sv = ScrollView(size=self.size, pos=self.pos)
			self.bind(size=self._set_sv_size, pos=self._set_sv_pos)
			layout = BoxLayout(orientation="vertical")
			self._sv.add_widget(layout)
			self.add_widget(self._sv)
		else:
			layout = self._sv.children[0]
			layout.clear_widgets()
		for txt, part in self.options:
			if not callable(part):
				raise TypeError("Menu options must be callable")
			butn = Button(
				text=txt,
				on_release=part,
				font_name=self.font_name,
				font_size=self.font_size,
				valign="center",
				halign="center",
			)
			butn.bind(size=butn.setter("text_size"))
			layout.add_widget(butn)


class Dialog(BoxLayout):
	"""MessageBox with a DialogMenu beneath it.

	Set the properties ``message_kwargs`` and ``menu_kwargs``,
	respectively, to control them -- but you probably want
	to do that by returning a pair of dicts from an action
	in lisien.

	"""

	message_kwargs = DictProperty({})
	menu_kwargs = DictProperty({})

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.bind(
			message_kwargs=self._propagate_msg_kwargs,
			menu_kwargs=self._propagate_menu_kwargs,
		)
		self._propagate_msg_kwargs()
		self._propagate_menu_kwargs()

	@partial(logwrap, section="Dialog")
	def _propagate_msg_kwargs(self, *_):
		if "msg" not in self.ids:
			Clock.schedule_once(self._propagate_msg_kwargs, 0)
			return
		kw = dict(self.message_kwargs)
		kw.setdefault(
			"background", "atlas://data/images/defaulttheme/textinput"
		)
		for k, v in kw.items():
			setattr(self.ids.msg, k, v)

	@partial(logwrap, section="Dialog")
	def _propagate_menu_kwargs(self, *_):
		if "menu" not in self.ids:
			Clock.schedule_once(self._propagate_menu_kwargs, 0)
			return
		kw = dict(self.menu_kwargs)
		kw.setdefault(
			"background",
			"atlas://data/images/defaulttheme/vkeyboard_background",
		)
		for k, v in kw.items():
			setattr(self.ids.menu, k, v)


class DialogLayout(FloatLayout):
	"""A layout, normally empty, that can generate dialogs

	To make dialogs, set my ``todo`` property to a list. It may contain:

	* Strings, which will be displayed with an "OK" button to dismiss them
	* Lists of pairs of strings and callables, which generate buttons with the
	  string on them that, when clicked, call the callable
	* Lists of pairs of dictionaries, which are interpreted as keyword
	  arguments to :class:`MessageBox` and :class:`DialogMenu`

	In place of a callable you can use the name of a function in my
	``usermod``, a Python module given by name. I'll import it when I need it.

	Needs to be instantiated with a lisien ``engine`` -- probably an
	:class:`EngineProxy`.

	"""

	dialog = ObjectProperty()
	todo = ListProperty()
	usermod = StringProperty("user")
	userpkg = StringProperty(None, allownone=True)

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.dialog = Dialog()
		self._finalize()

	@logwrap(section="DialogLayout")
	def _finalize(self, *_):
		app = App.get_running_app()
		if not hasattr(app, "engine"):
			Clock.schedule_once(self._finalize, 0)
			return
		engine = app.engine
		todo = engine.universal.get("last_result")
		if isinstance(todo, list):
			self.todo = todo
		else:
			self.todo = []
		self.idx = engine.universal.get("last_result_idx", 0)
		engine.universal.connect(self._pull)
		if self.todo:
			self.advance_dialog()

	@logwrap(section="DialogLayout")
	def _pull(self, *_, key, value):
		if key == "last_result":
			self.todo = value if value and isinstance(value, list) else []
		elif key == "last_result_idx":
			self.idx = value if value and isinstance(value, int) else 0

	@logwrap(section="DialogLayout")
	def on_idx(self, *_):
		lidx = self.engine.universal.get("last_result_idx")
		if lidx is not None and lidx != self.idx:
			self.engine.universal["last_result_idx"] = self.idx
		Logger.debug(f"DialogLayout.idx = {self.idx}")

	@mainthread
	@logwrap(section="DialogLayout")
	def advance_dialog(self, after_ok=None, *args):
		"""Try to display the next dialog described in my ``todo``.

		:param after_ok: An optional callable. Will be called after the user clicks
		a dialog option--as well as after any game-specific code for that option
		has run.

		"""
		self.clear_widgets()
		try:
			Logger.debug(f"About to update dialog: {self.todo[0]}")
			self._update_dialog(self.todo.pop(0), after_ok)
		except IndexError:
			if after_ok is not None:
				after_ok()

	@mainthread
	@logwrap(section="DialogLayout")
	def _update_dialog(self, diargs, after_ok, **kwargs):
		if diargs is None:
			Logger.debug("DialogLayout: null dialog")
			return
		if isinstance(diargs, tuple) and diargs[0] == "stop":
			if len(diargs) == 1:
				Logger.debug("DialogLayout: null dialog")
				return
			diargs = diargs[1]
		dia = self.dialog
		# Simple text dialogs just tell the player something and let them click OK
		if isinstance(diargs, str):
			dia.message_kwargs = {"text": diargs}
			dia.menu_kwargs = {
				"options": [("OK", partial(self.ok, cb=after_ok))]
			}
		# List dialogs are for when you need the player to make a choice and don't care much
		# about presentation
		elif isinstance(diargs, list):
			dia.message_kwargs = {"text": "Select from the following:"}
			dia.menu_kwargs = {
				"options": list(
					map(partial(self._munge_menu_option, after_ok), diargs)
				)
			}
		# For real control of the dialog, you need a pair of dicts --
		# the 0th describes the message shown to the player, the 1th
		# describes the menu below
		elif isinstance(diargs, tuple):
			if len(diargs) != 2:
				# TODO more informative error
				raise TypeError("Need a tuple of length 2")
			msgkwargs, mnukwargs = diargs
			if isinstance(msgkwargs, dict):
				dia.message_kwargs = msgkwargs
			elif isinstance(msgkwargs, str):
				dia.message_kwargs["text"] = msgkwargs
			else:
				raise TypeError("Message must be dict or str")
			if isinstance(mnukwargs, dict):
				mnukwargs["options"] = list(
					map(
						partial(self._munge_menu_option, after_ok),
						mnukwargs["options"],
					)
				)
				dia.menu_kwargs = mnukwargs
			elif isinstance(mnukwargs, (list, tuple)):
				dia.menu_kwargs["options"] = list(
					map(partial(self._munge_menu_option, after_ok), mnukwargs)
				)
			else:
				raise TypeError("Menu must be dict or list")
		else:
			raise TypeError(
				"Don't know how to turn {} into a dialog".format(type(diargs))
			)
		if dia.parent != self:
			Logger.debug("DialogLayout: Adding the dialog to the layout")
			self.add_widget(dia)
		else:
			Logger.debug("DialogLayout: Dialog is already in the layout")

	@logwrap(section="DialogLayout")
	def ok(self, *_, cb=None, cb2=None):
		"""Clear dialog widgets, call ``cb`` if provided, and advance the dialog queue

		``cb2`` will be called after advancing the dialog."""
		self.clear_widgets()
		if cb:
			cb()
		self.advance_dialog(after_ok=cb2)

	@logwrap(section="DialogLayout")
	def _lookup_func(self, funcname):
		from importlib import import_module

		if not hasattr(self, "_usermod"):
			self._usermod = import_module(self.usermod, self.userpkg)
		return getattr(self.usermod, funcname)

	@logwrap(section="DialogLayout")
	def _munge_menu_option(self, after_ok, option):
		if not isinstance(option, tuple):
			raise TypeError
		name, func = option
		if func is None:
			return name, partial(self.ok, cb2=after_ok)
		if callable(func):
			return name, partial(self.ok, cb=func, cb2=after_ok)
		if isinstance(func, tuple):
			fun = func[0]
			if isinstance(fun, str):
				fun = self._lookup_func(fun)
			args = func[1]
			if len(func) == 3:
				kwargs = func[2]
				func = partial(fun, *args, **kwargs)
			else:
				func = partial(fun, *args)
		if isinstance(func, str):
			func = self._lookup_func(func)
		return name, partial(self.ok, cb=func, cb2=after_ok)


store_kv(
	__name__,
	"""
<Box>:
	canvas.before:
		Color:
			rgba: self.background_color
		BorderImage:
			border: self.border
			pos: self.pos
			size: self.size
			source: self.background
		Color:
			rgba: 1, 1, 1, 1
<ScrollableLabel>:
	Label:
		size_hint_y: None
		height: self.texture_size[1]
		text_size: self.width, None
		text: root.text
		color: root.color
<MessageBox>:
	ScrollableLabel:
		x: root.x + root.padding[0]
		y: root.y + root.padding[3]
		width: root.width - root.padding[2]
		height: root.height - root.padding[1]
		text: root.text
		color: root.foreground_color
<Dialog>:
	orientation: 'vertical'
	pos_hint: {'x': 0, 'y': 0}
	size_hint: 1, 0.3
	MessageBox:
		id: msg
	DialogMenu:
		id: menu
""",
)
